/**
 * @file cpudll.c
 * This dll creates a cpu thread in which
 * cpu is running. Also handles correct cpu timing
 * and functionality of interrupt mechanism.
 */

//#define BENCHMARK

// Windows Header Files:
#include <windows.h>

// C RunTime Header Files
#include <stdlib.h>
#include <tchar.h>

#include "../dllkit/dllkit.h"
#include "cpu.h"
#include "../usrmsgs.h"



/** Frekvency of procesor */
UINT Freq = 1790000;
/** No. of instructions per milisecond */
UINT Frms = 1790;

/* ----- Forward declarations */

DWORD WINAPI CpuThreadFunc(LPVOID lpParam);
void CreateCpuThread();
void ExitCpuThread();
void ContExe();

/* ----- End of forward declarations */


UINT start;	/**< Start time of CPUs execution */
UINT total;	/**< No of ticks from @start */
UINT secstart;	/**< Helper variable used for determining and correcting the CPU speed */
UINT ticks;	/**< Last value of  GetTickCount(); */

typedef enum { NOTSTARTED, RUNNING, ENDED } THRSTATE;

BOOL volatile Running;	/**< Can be used by two threads! */
THRSTATE volatile ThreadState;	/**< Used for synchronization (to make sure, thread has correctly ended) */

typedef enum { NONE, IRQ, NMI, RESET } INTR;

INTR volatile Interrupt; /**< TRUE if interrupt occures */

HANDLE CpuThreadHandle;	/**< Handle of this thread */
DWORD CpuThreadID;	/**< ID of this thread */

HANDLE hThisModule;
MSGLIST Msgs;
INITDLLSTRUCT info;

/**
 * @brief Dll entry point
 */
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, 
					  LPVOID lpReserved)
{
	hThisModule = hModule;
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		break;
	case DLL_PROCESS_DETACH:
		break;
	}	
    return TRUE;
}
					
DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idi)
{
	Running = FALSE;
	ThreadState = NOTSTARTED;

	CreateMsgList(&Msgs, 11);
	AddMessage(Msgs, NMIMSG);
	AddMessage(Msgs, IRQMSG);
	AddMessage(Msgs, RESETMSG);
	AddMessage(Msgs, SETFREQ);
	AddMessage(Msgs, GETREGS);
	AddMessage(Msgs, ISRUNNING);
	AddMessage(Msgs, ISTEP);
	AddMessage(Msgs, WAITEND);
	AddMessage(Msgs, IRQD);
	AddMessage(Msgs, NMID);
	AddMessage(Msgs, RESETD);
	AddMessage(Msgs, STARTCPU);
	AddMessage(Msgs, PAUSECPU);

	info.CSMemory = ids->CSMemory;
	info.hInst = ids->hInst;
	info.hWnd = ids->hWnd;
	info.memory = ids->memory;
	info.memgetb = ids->memgetb;
	memgetb = ids->memgetb;
	info.memsetb = ids->memsetb;
	memsetb = ids->memsetb;
	info.optional = ids->optional;

	Interrupt = RESET;

	idi->iolo = 0;	/* No io memory needed */
	idi->iohi = 0;
	idi->msglist = Msgs;
}

/** Called from another thread!! */
DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds)
{
	switch (mds->message) {
	case STARTCPU:
		CreateCpuThread();
		SendMessage(info.hWnd, CPURUNNING, 0, 0);
		break;
	case PAUSECPU:
		ExitCpuThread();
		SendMessage(info.hWnd, CPUSTOPPED, 0, 0);
		break;
	case NMIMSG:
		Interrupt = NMI;
		break;
	case IRQMSG:
		Interrupt = IRQ;
		break;
	case RESETMSG:
		Interrupt = RESET;
		break;
	case SETFREQ:
		Freq = (UINT) mds->lParam;
		Frms = Freq / 1000;
		break;


	// Debugger messages:

	case WAITEND:
		// wait until CpuThread stops
		while (ThreadState != ENDED && ThreadState != NOTSTARTED)
			;
		break;
	case GETREGS:
		if (Running == TRUE) {
			*(Regs**)(mds->lParam) = NULL;
			break;
		}
		*(Regs**)(mds->lParam) = GetRegs();
		break;
	case ISTEP:
		if (Running == TRUE) {
			*(UINT*)(mds->lParam) = 0;
			break;
		}
		*(UINT*)(mds->lParam) = InstrStep();
		break;
	case IRQD:
		Irq();
		break;
	case NMID:
		Nmi();
		break;
	case RESETD:
		Reset();
		break;
	}
}

DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, 
										  iomode mode)
{
}

/**
 * @brief Creates a new CPU thread.
 * 
 * Creates a new thread in which processor is running.
 * Called by @a DllMain when loading this dll.
 * @see CpuThreadFunc
 * @see DllMain
 * @see ExitCpuThread
 */

void CreateCpuThread()
{
	secstart = start = GetTickCount();
	total = 0;
	Running = TRUE;
	CpuThreadHandle = CreateThread(NULL, 0, &CpuThreadFunc, NULL, 0, &CpuThreadID);
}

/**
 * @brief The main thread function
 *
 * This function is called by @a CreateCpuThread()
 * @param lpParam Value of this param doesn't matter
 * @return Nothing
 */
DWORD WINAPI CpuThreadFunc(LPVOID lpParam)
{
	ThreadState = RUNNING;
	while (Running) {
			ContExe();
	}
	ThreadState = ENDED;
	return 0;
}

/**
 * @brief Executes needed amount of instructions
 *
 * Executes appropriate amount of instructions according to given frequency.
 * @a secstart must be set by @a CreateCpuThread() when called the first time
 * @see CreateCpuThread()
 */
void ContExe()
{  
	UINT i;
	UINT todo = ((ticks = GetTickCount()) - secstart) * Frms;
	UINT tmp = 0;
	secstart = ticks + todo / Freq;

	for (i = 1; (i <= todo) && (Interrupt == NONE); i += tmp) {
		tmp = InstrStep();
		total = (total + tmp) % 0xFFFFFFF0;
	}
	if (Interrupt != NONE) {
		switch (Interrupt) {
		case IRQ:
			Irq();
			break;
		case NMI:
			Nmi();
			break;
		case RESET:
			Reset();
			break;
		}
	}
	Interrupt = NONE;
}

/**
 * @brief Kills this thread
 * @see CreateCpuThread()
 */
void ExitCpuThread()
{
	Running = FALSE;
}