/**
 * @file debugdll.c
 * Simple debugger & disassembler for emu6502
 */

#include "../dllkit/dllkit.h"
#include "../usrmsgs.h"
#include "resource.h"
#include "instructions.h"
#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <stdio.h>

#define ADR(X, Y) (((Y) << 8) + (X))

#define FN 0x80		/**< Negative flag */
#define FV 0x40		/**< Overflow flag */
#define FB 0x10		/**< BRK flag */
#define FD 0x08		/**< Decimal flag */
#define FI 0x04		/**< Interrupt flag */
#define FZ 0x02		/**< Zero flag */
#define FC 0x01		/**< Caryy flag */


typedef struct {
	UINT PC;			/**< Program counter */
	unsigned char X;	/**< Index register X */	
	unsigned char Y;	/**< Index register Y */
	unsigned char A;	/**< Accumulator */
	unsigned char P;	/**< Status register */
	UINT S;				/**< Stack pointer */
} Regs;

HANDLE hThisModule;
MSGLIST Msgs;
INITDLLSTRUCT info;

/** @brief Handle to the main menu bar */
HMENU hMenu;
/** @brief Handle to local debug menu which is appended to @a hMenu */
HMENU hDebugMenu;

/* ----- begin forward declarations */

/** 
 * @brief Starts debug
 *
 * Retrieves pointer to registers from the cpu plugin, if not present,
 * shows an error message and stops debugging .
 */
void StartDebug();

/**
 * @brief Disassembler
 * 
 * Disassembles given range of memory and writes it into the stream specified
 *
 * @param lower, upper Range of memory to be disassembled
 * @param fil Handle to stream to write disassembled data into
 * @param act Adress of instruction to be highlighted by adding >>> before it
 */
void DisAsm(int lower, int upper, HANDLE fil, int act);


/**
 * @brief Printer of memory in hex format
 *
 * Prints given range of memory into given stream in hexadecimal format
 *
 * @param adr1, adr2 Range of memory to be printed
 * @param fil Stream to write the memory into
 */
void DumpMemHex(HANDLE fil, UINT adr1, UINT adr2);

/**
 * @brief Printer of memory in raw format
 *
 * Prints given range of memory into given stream in raw format
 *
 * @param adr1, adr2 Range of memory to be printed
 * @param fil Stream to write the memory into
 */
void DumpMem(HANDLE fil, UINT adr1, UINT adr2);

/** @brief Displays usage instructions */
void printhelp();

/**
 * @brief Shows save dialog 
 *
 * Shows save dialog and opens selected file.
 *
 * @return Handle to just opened stream of the selected file
 */
HANDLE ofile();

/**
 * @brief Shows open dialog 
 *
 * Shows open dialog and opens selected file.
 *
 * @return Handle to just opened stream of the selected file
 */
HANDLE orfile();

/**
 * @brief Helper procedure for @a StartDebug
 *
 * Parses registers modifying commands.
 *
 * @param regs Pointer to the Regs structure used by the CPU plugin
 * @param buf Pointer to the command user has entered
 */
void SetReg(Regs* regs, char* buf);

/** @brief Prints actual state of registers onto console */
void printregs(Regs* regs);

/**
 * @brief Prints message onto screen or into specified stream.
 *
 * If the fil stream is NULL, the message will be printed onto console
 *
 * @param fil Handle to a stream or NULL
 * @param str Message to be printed
 */
void printfil(HANDLE fil, char* str);

/** @brief Reads an address interval and controls if it is correct */
void rinterval(UINT* adr1, UINT* adr2);

/** @brief Reads an hexadecimal integer and controls if it is correct */
UINT rhex(char* msg);

/** @brief Converts strings read from console to null terminated strings */
void mklpstr(char* str);

/** @brief Prints given message onto the console */
void print(char* str);

/** @brief Reads one command from console */
void rline(char* str);

/** @brief Prints registers and dissassembled position in memory */
void printact(Regs* regs);

/* ----- end forward declarations */
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, 
					  LPVOID lpReserved)
{
	hThisModule = hModule;
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		CreateMsgList(&Msgs, 1);
		break;
	case DLL_PROCESS_DETACH:
		FreeMsgList(Msgs);
		break;
	}	
    return TRUE;
}

DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idi)
{
	TBADDBITMAP tbab;
	TBBUTTON tbt;
	int index;

	// Add Message list of messages to be catched
	AddMessage(Msgs, WM_COMMAND);

	// Copy informations passed to this dll
	info.CSMemory = ids->CSMemory;
	info.hInst = ids->hInst;
	info.hWnd = ids->hWnd;
	info.memory = ids->memory;
	info.memgetb = ids->memgetb;
	info.memsetb = ids->memsetb;
	info.optional = ids->optional;
	info.hTb = ids->hTb;

	hMenu = GetMenu(info.hWnd);
	hDebugMenu = LoadMenu(hThisModule, MAKEINTRESOURCE(IDR_DEBUG));
	AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT) hDebugMenu, "&Debug");
	DrawMenuBar(info.hWnd);

	tbab.hInst = hThisModule;
	tbab.nID = IDR_TB;
	index = SendMessage(info.hTb, TB_ADDBITMAP, (WPARAM) 1, 
        (WPARAM) &tbab); 
	ZeroMemory(&tbt, sizeof(TBBUTTON));
	tbt.iBitmap = index;
	tbt.idCommand = ID_DEBUG_STARTDEBUG;
	tbt.fsState = TBSTATE_ENABLED;
	tbt.fsStyle = TBSTYLE_BUTTON | TBSTYLE_AUTOSIZE;

    SendMessage(info.hTb, TB_ADDBUTTONS, (WPARAM) 1, 
        (LPARAM) &tbt); 
    SendMessage(info.hTb, TB_AUTOSIZE, 0, 0); 

   
	// We don't need any io memory
	idi->iolo = 0;
	idi->iohi = 0;
	idi->msglist = Msgs;
}


DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds)
{
	int wmId, wmEvent;

	switch (mds->message) {
	case WM_COMMAND:
		wmId = LOWORD(mds->wParam); 
		wmEvent = HIWORD(mds->wParam); 
		switch (wmId) {
		case ID_DEBUG_STARTDEBUG:
			ShowWindow(info.hWnd, SW_MINIMIZE);
			// Stop CPU
			SendMessage(info.hWnd, PAUSECPU, 0, 0);
			SendMessage(info.hWnd, WAITEND, 0, 0);
			AllocConsole();
			StartDebug();
			FreeConsole();
			ShowWindow(info.hWnd, SW_RESTORE);
			break;
		case ID_DEBUG_ABOUT:
			MessageBox(info.hWnd, "Debug plugin\nAdds some debugging & disassembling features to emu6502\n(c)2002 Petr Cermak", "About", MB_OK);
			break;
		}
		break;
	}

}

DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, 
										  iomode mode)
{
}

void StartDebug()
{
    char buf[256]; 
	UINT adr1, adr2, val;
	DWORD tmp;
	Regs* regs = NULL;
	HANDLE fil = NULL;

	print("Emu6502 debugger & disassembler\n");
	print("To retrieve list of commands, type 'h' and press enter.\n");

	// Retrieve pointer to the Regs struct.
	SendMessage(info.hWnd, GETREGS, 0, (LPARAM) &regs);

	if (regs == NULL) {
		print("Cpu plugin not running!\nDebugging not available.\nPress Enter key to return.\n");
		rline(buf);
		return;
	}

	do {
		print("\n");
		print("> ");
		rline(buf);
		switch (buf[0]) {
		case 'H':
		case 'h':
			printhelp();
			break;
		case 'D':
		case 'd':
			print("Disassemble\n");
			fil = NULL;
			rinterval(&adr1, &adr2);
			if (buf[1] == 'f') {
				fil = ofile();
				if (fil == NULL) {
					break;
				}
			}
			DisAsm(adr1, adr2, fil, -1);
			if (fil != NULL) {
				CloseHandle(fil);
			}
			break;
		case 'M':
		case 'm':
			fil = NULL;
			if (buf[1] == 'f' || buf[1] == 'F') {
				rinterval(&adr1, &adr2);
				if (buf[2] == 'h' || buf[2] == 'H') {
					fil = ofile();
					if (fil == NULL) {
						break;
					}
					DumpMemHex(fil, adr1, adr2);
				}
				else {
					fil = ofile();
					if (fil == NULL) {
						break;
					}
					DumpMem(fil, adr1, adr2);
				}	
				CloseHandle(fil);
			}
			else if (buf[1] == 's' || buf[1] == 'S'){
				if(buf[2] == 'f' || buf[2] == 'F'){
					do {
						adr1 = rhex("From addr.: ");
					} while (adr1 > 0xFFFF);
					fil = orfile();
					if (fil == NULL) {
						break;
					}
					adr2 = GetFileSize(fil, NULL);
					if (adr1 + adr2 - 1 > 0xFFFF) {
						print("Given address and size overreach the $ffff boundary\n");
						CloseHandle(fil);
						break;
					}
					if (!ReadFile(fil, info.memory + adr1, adr2, &tmp, NULL)) {
						print("Error reading file\n");
						break;
					}
				} 
				else {
					rinterval(&adr1, &adr2);
					do {
						val = rhex("Value: ");
					} while (val > 0xFF);
					for (tmp = adr1; tmp <= adr2; tmp++) {
						(*info.memsetb)(tmp, (unsigned char) val);
					}
				}
			}
			else {
				rinterval(&adr1, &adr2);
				DumpMemHex(fil, adr1, adr2);
			}
			break;
		case 'R':
		case 'r':
			printregs(regs);
			break;
		case 'A':
		case 'a':
			printact(regs);
			break;
		case 'N':
		case 'n':
			SendMessage(info.hWnd, ISTEP, 0, (LPARAM) &adr1);
			printact(regs);
			sprintf(buf, "Instruction took %d cycles\n", adr1);
			print(buf);
			break;
		case 'I':
		case 'i':
			switch(buf[1]) {
			case 'I':
			case 'i':
				print("IRQ\n");
				(*info.memsetb)(IRQID, 254);
				SendMessage(info.hWnd, IRQD, 0, 0);
				printact(regs);
				break;
			case 'N':
			case 'n':
				print("NMI\n");
				(*info.memsetb)(NMIID, 254);
				SendMessage(info.hWnd, NMID, 0, 0);
				printact(regs);
				break;
			case 'R':
			case 'r':
				print("RESET\n");
				(*info.memsetb)(RESETID, 254);
				SendMessage(info.hWnd, RESETD, 0, 0);
				printact(regs);
				break;
			default:
				print("Unknown interrupt name\n");
				break;
			}
			break;
		case 'S':
		case 's':
			SetReg(regs, buf);
			break;
		case 'Q':
		case 'q':
			break;
		default:
			print("Unknown command, type 'h' for help\n");
			break;
		}
	} while (buf[0] != 'q');
}

void SetReg(Regs* regs, char* buf)
{
	UINT adr1;
	BOOL ok = FALSE;
	
	if (buf[1] != 'f' && buf[1] != 'F') {
		printregs(regs);
	}

	do {
		ok = FALSE;
		switch(buf[1]) {
		case 'A':
		case 'a':
			adr1 = rhex("A: ");
			if (adr1 <= 255) {
				ok = TRUE;
				regs->A = adr1;
			}
			break;
		case 'X':
		case 'x':
			adr1 = rhex("X: ");
			if (adr1 <= 255) {
				ok = TRUE;
				regs->X = adr1;
			}
			break;
		case 'Y':
		case 'y':
			adr1 = rhex("Y: ");
			if (adr1 <= 255) {
				ok = TRUE;
				regs->Y = adr1;
			}
			break;
		case 'S':
		case 's':
			adr1 = rhex("S: ");
			if (adr1 >= 256 && adr1 <511) {
				ok = TRUE;
				regs->A = adr1;
			}
			break;
		case 'P':
		case 'p':
			if (buf[2] == 'c' || buf[2] == 'C') {
				adr1 = rhex("PC: ");
				if (adr1 <= 65535) {
					ok = TRUE;
					regs->PC = adr1;
				}
			} 
			else {
				adr1 = rhex("P: ");
				if (adr1 <= 255) {
					ok = TRUE;
					regs->P = adr1;
				}
			}
			break;
		case 'F':
		case 'f':
			ok = TRUE;
			switch(buf[2]) {
			case 'N':
			case 'n':
				regs->P = regs->P ^ FN;
				break;
			case 'V':
			case 'v':
				regs->P = regs->P ^ FV;
				break;
			case 'B':
			case 'b':
				regs->P = regs->P ^ FB;
				break;
			case 'D':
			case 'd':
				regs->P = regs->P ^ FD;
				break;
			case 'I':
			case 'i':
				regs->P = regs->P ^ FI;
				break;
			case 'Z':
			case 'z':
				regs->P = regs->P ^ FZ;
				break;
			case 'C':
			case 'c':
				regs->P = regs->P ^ FC;
				break;
			default:
				print("Wrong name of flag\n");
				break;
			}
			break;
		default:
			print("Wrong name of register\n");
			ok = TRUE;
			break;
		}
	} while (ok == FALSE);
	printregs(regs);
}


void DisAsm(int lower, int upper, HANDLE fil, int act)
{
	int adr = lower;
	int tmp;
	unsigned char instr, op1, op2;
	char buf[20];

	while (adr <= upper) {
		instr = (*info.memgetb)(adr);
		if (act == adr) {
			printfil(fil, ">>> ");
		}
		sprintf(buf, "%04x\t", adr);
		printfil(fil, buf);
		printfil(fil, INSTRUCTIONS[instr].name);

		switch (INSTRUCTIONS[instr].aType) {
		case IMPL:
			break;
		case ACC:
			printfil(fil, " A");
			break;
		case IMM:
			printfil(fil, " #$");
			op1 = (*info.memgetb)(adr + 1);
			sprintf(buf, "%02x", op1);
			printfil(fil, buf);
			break;
		case ABS:
			printfil(fil, " $");
			op1 = (*info.memgetb)(adr + 1);
			op2 = (*info.memgetb)(adr + 2);
			sprintf(buf, "%04x", ADR(op1, op2));
			printfil(fil, buf);
			break;
		case ABSX:
			printfil(fil, " $");
			op1 = (*info.memgetb)(adr + 1);
			op2 = (*info.memgetb)(adr + 2);
			sprintf(buf, "%04x, X", ADR(op1, op2));
			printfil(fil, buf);
			break;
		case ABSY:
			printfil(fil, " $");
			op1 = (*info.memgetb)(adr + 1);
			op2 = (*info.memgetb)(adr + 2);
			sprintf(buf, "%04x, Y", ADR(op1, op2));
			printfil(fil, buf);
			break;
		case ZP:
			printfil(fil, " $");
			op1 = (*info.memgetb)(adr + 1);
			sprintf(buf, "%02x", op1);
			printfil(fil, buf);
			break;
		case ZPX:
			printfil(fil, " $");
			op1 = (*info.memgetb)(adr + 1);
			sprintf(buf, "%02x, X", op1);
			printfil(fil, buf);
			break;
		case ZPY:
			printfil(fil, " $");
			op1 = (*info.memgetb)(adr + 1);
			sprintf(buf, "%02x, Y", op1);
			printfil(fil, buf);
			break;
		case REL:
			op1 = (*info.memgetb)(adr + 1);
			printfil(fil, " $");
			if (op1 > 0x7F) {
				tmp = adr + 2 - 0x100 + op1;
			}
			else {
				tmp = adr + 2 + op1;
			}
			sprintf(buf, "%04x", tmp);
			printfil(fil, buf);
			break;
		case IND:
			printfil(fil, " (");
			op1 = (*info.memgetb)(adr + 1);
			op2 = (*info.memgetb)(adr + 2);
			sprintf(buf, "%04x)", ADR(op1, op2));
			printfil(fil, buf);
			break;
		case INDX:
			printfil(fil, " (");
			op1 = (*info.memgetb)(adr + 1);
			sprintf(buf, "%02x, X)", op1);
			printfil(fil, buf);
			break;
		case INDY:
			printfil(fil, " (");
			op1 = (*info.memgetb)(adr + 1);
			sprintf(buf, "%02x), Y", op1);
			printfil(fil, buf);
			break;
		}
		printfil(fil, "\r\n");
		adr += INSTRUCTIONS[instr].len;
	}
}

void printhelp()
{
	print("(a)ctual state   --   prints actual state of emulator (registers and disassembled position in memory)\n");
	print("(d)isassemble   --   disassembles given range of memory\n");
	print("(d)isassemble into (f)ile\n");
	print("(i)nterrupt (i)rq   --   raises interrupt request\n");
	print("(i)nterrupt (n)mi   --   raises nonmaskable interrupt\n");
	print("(i)nterrupt(r)eset   --   raises reset interrupt\n");
	print("(h)elp   --   this message\n");
	print("(m)emory dump   --   prints memory dump in hex format onto screen\n");
	print("(m)emory dump into (f)ile   --   dumps given range of memory into file (binary)\n");
	print("(m)emory dump into (f)ile in (h)ex format\n");
	print("(m)emory (s)et   --   set given range of memory to specified value\n");
	print("(m)emory (s)et from (f)ile   --   load block of memory from file\n");
	print("(n)ext instruction   --   executes one instruction\n");
	print("(q)uit debugger\n");
	print("(r)egisters   --   prints actual state of registers\n");
	print("(s)et ([A, X, Y, P, PC, S]) register to given value\n");
	print("(s)et (f)lag ([N, V, B, D, I, Z, C])   --   negates given flag\n");
	print("Creating commands - example:\n");
	print("(m)emory dump\tinto (f)ile in (h)ex format\n");
	print("means: > mfh <enter>\n");
	print("All values must be in hex format!\n");
}

void DumpMem(HANDLE fil, UINT adr1, UINT adr2)
{
	DWORD tmp;
	WriteFile(fil, info.memory + adr1, adr2 - adr1 + 1, &tmp, NULL);
}


void DumpMemHex(HANDLE fil, UINT adr1, UINT adr2)
{
	UINT i;
	char buf[10];
	for (i = adr1; i <= adr2; i++) {
		if ((i - adr1) % 20 == 0) {
			sprintf(buf, "\r\n%04x    ", i);
			printfil(fil, buf);
		}
		sprintf(buf, "%02x ", (*info.memgetb)(i));
		printfil(fil, buf);
	}
}

HANDLE orfile()
{
	OPENFILENAME ofn;
	TCHAR FileName[1024];
	HANDLE hFile;

	lstrcpy(FileName, "");
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.lpstrFile = FileName;
	ofn.hwndOwner = info.hWnd;
	ofn.nMaxFile = sizeof(FileName);
	ofn.lpstrFilter = 
		TEXT("Pratial memory image files (*.bin;*.img)\0*.bin;*.img\0All Files (*.*)\0*.*\0");
	ofn.nFilterIndex = 1;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
	
	ShowWindow(info.hWnd, SW_RESTORE);
	if (!GetOpenFileName(&ofn))
		return NULL;
	ShowWindow(info.hWnd, SW_MINIMIZE);
	
	hFile=CreateFile((LPCSTR) FileName, GENERIC_READ, 0,NULL, OPEN_EXISTING, 0, NULL);
	if (hFile==INVALID_HANDLE_VALUE) {
		print("Cannot open file\n");
		return NULL;
	}
	return hFile;
}



HANDLE ofile()
{
	OPENFILENAME ofn;
	TCHAR FileName[1024];
	HANDLE hFile;

	lstrcpy(FileName, "");
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.lpstrFile = FileName;
	ofn.hwndOwner = info.hWnd;
	ofn.nMaxFile = sizeof(FileName);
	ofn.lpstrFilter = 
		TEXT("All Files (*.*)\0*.*\0");
	ofn.nFilterIndex = 1;
	ofn.Flags = OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST;
	
	ShowWindow(info.hWnd, SW_RESTORE);
	if (!GetSaveFileName(&ofn)) {
		ShowWindow(info.hWnd, SW_MINIMIZE);
		return NULL;
	}
	ShowWindow(info.hWnd, SW_MINIMIZE);

	hFile=CreateFile((LPCSTR) FileName, GENERIC_WRITE, 0,NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile==INVALID_HANDLE_VALUE) {
		print("Error creating file\n");
		return FALSE;
	}
	return hFile;
}


void printact(Regs* regs)
{
	printregs(regs);
	print("\n");
	DisAsm(regs->PC,
		((regs->PC + 10) > 0xffff) ? 0xffff : (regs->PC + 10), NULL, regs->PC);
}

#define GETFLAG(FL)	\
	((regs->P & (FL)) == 0) ? 0 : 1

void printregs(Regs* regs)
{
	char buf[255];
	sprintf(buf, "PC = %04x, S = %04x\nA = %02x, X = %02x, Y = %02x\n", 
		regs->PC, regs->S, regs->A, regs->X, regs->Y);
	print(buf);
	sprintf(buf, "P = %02x: N = %d, V = %d, B = %d, D = %d, I = %d, Z = %d, C = %d\n", 
		regs->P, GETFLAG(FN), GETFLAG(FV), GETFLAG(FB), GETFLAG(FD), GETFLAG(FI), GETFLAG(FZ), GETFLAG(FC));
	print(buf);
}

void rinterval(UINT* adr1, UINT* adr2)
{
	do {
		*adr1 = rhex("From addr.: ");
		*adr2 = rhex("To addr.: ");
	} while (!(adr1 <= adr2 <= 0xFFFF));
}

UINT rhex(char* msg)
{
	char str[255];
	long num;
	str[0] = '0';
	str[1] = 'x';
	do {
		print(msg);
		rline(str + 2);
		print("\n");
		if (str[3] == 13 && str[2] == '0') {
			num = 0;
			break;
		}
	} while (!(num = strtoul(str, NULL, 16)));
	return (UINT) num;
}

void mklpstr(char* str)
{
	for(; *str != 13; str++)
		;
	*str = '\0';
}

void printfil(HANDLE fil, char* str)
{
	DWORD tmp;
	if (fil == NULL) {
		print(str);
	}
	else {
		WriteFile(fil, str, lstrlen(str), &tmp, NULL);
	}
}

void print(char* str)
{
	DWORD tmp;
	WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), str, lstrlen(str), &tmp, NULL);
}

void rline(char* str)
{
	DWORD tmp;
	ReadFile(GetStdHandle(STD_INPUT_HANDLE), str, 255, &tmp, NULL);
}