/**
 * @file dlluser.h
 * Include this file if you want to use emu6502 plugin dlls.
 */
#ifndef DLLUSER_H
#define DLLUSER_H

#include <windows.h>

/**
 * @brief List of messages
 *
 * Generic list of messages, used for specifying messages to be catched
 * or to be sent. Length of this list should be the number of messages
 * to be held within + 1, and the list should end with @a ENDOFMSGLIST
 */
typedef UINT* MSGLIST;

/** @brief Marks the end of the list of messages @a MSGLIST */
#define ENDOFMSGLIST 0xBFFF

/** @brief Structure containing various informations returned by @a InitDll */
typedef struct {
	MSGLIST msglist;	/**< @see MSGLIST */
	UINT iolo;	/**< Lower boundary of io memory reserved by this plugin */
	UINT iohi;  /**< Higher boundary of io memory reserved by this plugin */
} INITDLLINFO;


/**
 * @brief Initial informations passed to plugin when initializing.
 * 
 * Structure containing informations about the main application,
 * about 6502 memory and other.
 */
typedef struct {
	/** Pointer to shared memory. Use @a CSMemory when accessing it. */
	char* memory;
	/** Handle of the main window */
	HWND hWnd;
	/** Handle of the main window toolbar */
	HWND hTb;
	/** Instance of the emu6502 application */
	HINSTANCE hInst;
	/** Use when accesing shared @a memory */
	CRITICAL_SECTION CSMemory;
	/** Pointer to a function which gets a byte from given adress. 
	 * No need for synchronizing using @a CSMemory.*/
	unsigned char (*memgetb) (UINT adr);
	/** Pointer to a function which sets byte on given adress. 
	 * No need for synchronizing using @a CSMemory.*/
	void (*memsetb) (UINT adr, unsigned char byte);
	/** Area of allowed painting/refreshing */
	RECT rcPaint;
	/** Optional data passed by emu6502. */
	void* optional;
} INITDLLSTRUCT;


/** 
 * @brief When registered message occures, this message structure 
 * is passed to plugins.
 */
typedef struct {
	UINT message;
	WPARAM wParam;
	LPARAM lParam;
} MESSAGEDLLSTRUCT;

/** @brief Determines, if iofunc is called when reading or writing */
typedef enum { IOREAD, IOWRITE } iomode;

/**
 * @brief IO function mapped into some adress in memory
 *
 * Called directly when accesing adress to which is this function bound
 * @param adr Adress being accessed
 * @param data Byte being read/written
 * @param mode see @a iomode
 * @see iomode
 * @see memio
 */
typedef void (*pIOFunc)(UINT adr, unsigned char* data, iomode mode);

/** @brief emu6502 plugin dll initialization routine pointer */
typedef void (*pInitDllProc)(INITDLLSTRUCT*, INITDLLINFO*);

/** @brief emu6502 plugin dll message handling routine pointer */
typedef void (*pForwardDllMsgProc)(MESSAGEDLLSTRUCT*);


#endif	//#ifndef DLLUSER_H