/**
 * @file emu6502.h
 * Header file for @a emu6502.c, contains only local forward and type
 * declarations, not for reuse.
 */


#ifndef EMU6502_H
#define EMU6502_H

#include "resource.h"

#include "mem.h"
//#include "cpu.h"
//#include "cputhread.h"
#include "dlluser.h"
#include "usrmsgs.h"

/**
 * @brief Generic list of pointers
 */
typedef struct _PtrList PtrList;
struct _PtrList {
	void* ptr;
	PtrList* next;
};

/**
 * @brief List of messages and theirs catching dll functions.
 *
 * @see DllFuncNode
 * @see ForwardDllMsgProc
 */
typedef struct _DllMsgList DllMsgList;
struct _DllMsgList {
	UINT MesgID;		/**< ID of message to be catched */
	PtrList* Functions;	/**< List of functions catching this message */
	DllMsgList* next;	/**< Pointer to next member of this list */
};
	
/** @brief Entry point of this app. */
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
					 LPSTR lpCmdLine, int nCmdShow);
/** @brief Registers the main window class */
ATOM MyRegisterClass(HINSTANCE hInstance);

/**
 * @brief Inits & shows the main window
 *
 * Creates the main window, sizes it to fit the 256x256 canvas, creates
 * toolbar by calling @CreateToolbar, and then loads plugins by calling
 * @a LoadPlugins.
 *
 * @see CreateToolbar, LoadPlugins
 */
BOOL InitInstance(HINSTANCE, int);

/**
 * @brief Creates the main toolbar
 *
 * Creates main toolbar and adds one Open button to it
 *
 * @param hParentWnd Handle to parent window
 * @return Hanlde to the toolbar
 */
HWND CreateToolBar(HWND hParentWnd);

/**
 * @brief Main message handling routine
 *
 * Standard message processing routine, also forwards registered messages
 * to relevant plugins by calling @a ForwardDllMessage
 *
 * @see ForwardDllMessage
 */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

/** @brief About dialog's DialogProc */
LRESULT CALLBACK About(HWND, UINT, WPARAM, LPARAM);

/** 
 * @brief Loads memory image
 *
 * Stops execution of the CPU, loads memory image, also controls it's length
 */
BOOL LoadMemFile(HWND hWnd);

/**
 * @brief Error reporting proc.
 *
 * Shows a messagebox with given error mesage (from resources).
 * This message can also contain %s in this case the parameter name
 * becomes effective and prints on the position of %s.
 *
 * @param hWnd Handle of the parent window
 * @param uID Resource ID of the error message
 * @param name Informations to be printed on the %s position
 */
void error(HWND hWnd, UINT uID, TCHAR* name);

/**
 * @brief Loads plugins named in plugins.txt
 *
 * Goes through the plugins.txt file and loads named plugins by calling @a LoadPlugin procedure.
 * @param list Pointer to a @a DllMsgList list in which will be message IDs and theirs
 * catching procedures stored.
 * @param InstList List of handles to plugin-dlls loaded
 * @see LoadPlugin, ForwardDllMessage
 */
void LoadPlugins(DllMsgList** list, PtrList** InstList);

/**
 * @brief Loads plugin of given name.
 *
 * Executes his InitDll functions, receives list of messages to be catched
 * in this plugin and attaches it to the @a DllMsgList list.
 * @param name Name of plugin dll to be loaded.
 * @param list Pointer to a @a DllMsgList list in which will be message IDs and theirs
 * catching procedures stored.
 * @param InstList List of handles to plugin-dlls loaded
 * @see LoadPlugins, ForwardDllMessage
 */
BOOL LoadPlugin(TCHAR* name, DllMsgList** list, PtrList** InstList);

/*
 *
 */

void UnloadPlugins(PtrList* list);
/**
 * @brief Forwards message to registered plugin OnMessage functions.
 *
 * @param message ID of forwarded message
 * @param wParam wParam parameter of this message
 * @param lParam lParam parameter of this message
 * @param list List of messages and its catching functions. See @a DllMsgList
 * @return TRUE if relevant function found, FALSE otherwise
 */
BOOL ForwardDllMessage(UINT message, WPARAM wParam, LPARAM lParam, DllMsgList* list);

/**
 * @brief Adds message catching dll function to a dll message list.
 *
 * @param list pointer to a @a DllMsgList list
 * @param message ID of message to be catched
 * @param FDMProc Function, to which should be the message forwarded
 * Mostly function from plugin-dll.
 * @see ForwardDllMsgProc
 * @see DllMsgList
 */
void AddDllMsg(DllMsgList** list, UINT message, pForwardDllMsgProc FDMProc);

/**
 * @brief Allocates and adds new member to existing or NULL list
 */
void AddToPtrList(void* node, PtrList** list);

/**
 * @brief Helper function for LoadPlugins
 *
 * Sideeffect - makes buf to point to the next item
 * @param buf TCHAR buffer containing list of items separated by CRLF
 * and ended with \0.
 * @return first item of the list or NULL if there isn't any more
 * @see LoadPlugins
 */
TCHAR* GetNextStrToken(TCHAR** buf);

#endif // #ifndef EMU6502_H
