/**
 * @file videodll.c
 * A display plugin for emu6502. Reads video part of memory and 
 * displays what it contains. Also paints the text mode screen
 * with given charset.
 */

/**
 * @todo Crashes - send message to msg. loop and stop
 */

#include "../dllkit/dllkit.h"
#include "../usrmsgs.h"
#include "resource.h"
#include <stdio.h>
#include <tchar.h>

/** @name Timer defs */
/*@{*/
/** @brief ID of local timer */
#define MYTIMER 10001 
/** @brief The period between refreshing screen in ms.*/
#define PERIOD 50
/*@}*/

/** @name Screen defs */
/*@{*/
/** @brief Pixels per character */
#define FWIDTH 8
/** @brief Pixels per character */
#define FHEIGHT 8
/** @brief Characters per screen */
#define YCOUNT 32
/** @brief Characters per screen */
#define XCOUNT 32
/** @brief Width of the screen */
#define WIDTH FWIDTH * XCOUNT
/** @brief Height of the screen */
#define HEIGHT FHEIGHT * YCOUNT
/*@}*/

/** @name IO adresses */
/*@{*/
#define GRMODE IOLO + 0
#define LOADR IOLO + 1
#define HIADR IOLO + 2
#define LOCHR IOLO + 3
#define HICHR IOLO + 4
#define FGCOLR IOLO + 5
#define FGCOLG IOLO + 6
#define FGCOLB IOLO + 7
#define BKCOLR IOLO + 8
#define BKCOLG IOLO + 9
#define BKCOLB IOLO + 10
#define VIDEOON IOLO + 14
#define COMMAND IOLO + 15
/*@}*/

/** @name IO registers - copied after COMMAND */
/*@{*/
unsigned char volatile _GrMode = 0;
unsigned char volatile _LoAdr = 0x00;
unsigned char volatile _HiAdr = 0xA0;
unsigned char volatile _LoChr = 0x00;
unsigned char volatile _HiChr = 0xC0;
unsigned char volatile _BkColR = 0xFF;
unsigned char volatile _BkColG = 0xFF;
unsigned char volatile _BkColB = 0xFF;
unsigned char volatile _FgColR = 0x00;
unsigned char volatile _FgColG = 0x00;
unsigned char volatile _FgColB = 0x00;
/* */
unsigned char volatile GrMode = 0;
unsigned char volatile LoAdr = 0x00;
unsigned char volatile HiAdr = 0xA0;
unsigned char volatile LoChr = 0x00;
unsigned char volatile HiChr = 0xC0;
unsigned char volatile VideoOn = 0;
/*@}*/

#define ADR(X, Y) (((X) << 8) + (Y))

/** Initial begin of videomem (can change) */
/** @brief adress of videomem */
UINT memStart = 0xA000;
/** @brief adress of charmap */
UINT memChrStart = 0xC000;


/** @brief DIB header */
struct b4iheader {
	BITMAPV4HEADER bmiHeader; 
	RGBQUAD bmiColors[2];
} bGrInfo;

/** @brief An off-screen buffer for text mode */
char* OffScreen;

/** @brief Handle to the logo bitmap */
HBITMAP hLogo;
/** @brief A compatible DC used for drawing the logo */
HDC hOffDC;


/** @name Standard plugin variables */
/*@{*/
HANDLE hThisModule;
MSGLIST VideoMsgs;
volatile BOOL Running = FALSE;
INITDLLSTRUCT info;
/*@}*/


/* ----- begin forward declarations */

/** @brief Inits DIB and logo structures */
void InitGraphics();

/** @brief Paints the screen every n ms if needed */
VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime);

/** @brief Invalidates whole screen and evokes WM_PAINT */
void Repaint();

/**
 * @brief Paints whole screen.
 *
 * Called by WM_PAINT if the window needs repaint or every n ms when emulator is running
 * For creating text mode screen calls @a CopyChar.
 *
 * @param hdc DC of the canvas
 * @see CopyChar
 */
void PaintScreen(HDC hdc); 

/**
 * @brief Copies one character to given coordinates
 *
 * Uses character map stored in emulators memory
 *
 * @param dest Address of text mode videomem
 * @param adr Address of the character 
 * @param x X coordinate to paint the character onto
 * @param y Y coordinate to paint the character onto
 * @param invert Determines if the character bitmap should be inverted.
 * If so, character no. - 128 is used.
 */
void CopyChar(char* dest, UINT adr, int x, int y, BOOL invert);

/* ----- end forward declarations */

BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, 
					  LPVOID lpReserved)
{
	hThisModule = hModule;
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
		CreateMsgList(&VideoMsgs, 4);
		OffScreen = malloc(WIDTH * HEIGHT * sizeof(char) / 8);
		hLogo = LoadBitmap(hThisModule, MAKEINTRESOURCE(IDB_LOGO));
		VideoOn = FALSE;
		break;
	case DLL_PROCESS_DETACH:
		Running = FALSE;
		KillTimer(info.hWnd, MYTIMER);
		DeleteObject(hLogo);
		DeleteDC(hOffDC);
		free(OffScreen);
		/** @todo free VideoMsgs */
		break;
	}	
    return TRUE;
}

DLLEXPORT void __cdecl InitDll(INITDLLSTRUCT* ids, INITDLLINFO* idi)
{
	/* Add Message list of messages to be catched */
	AddMessage(VideoMsgs, WM_PAINT);
	AddMessage(VideoMsgs, CPURUNNING);
	AddMessage(VideoMsgs, CPUSTOPPED);
	AddMessage(VideoMsgs, STOPVIDEO);

	/* Copy informations passed to this dll */
	info.CSMemory = ids->CSMemory;
	info.hInst = ids->hInst;
	info.hWnd = ids->hWnd;
	info.memory = ids->memory;
	info.memgetb = ids->memgetb;
	info.memsetb = ids->memsetb;
	info.optional = ids->optional;
	info.rcPaint = ids->rcPaint;

	/* Initialize graphics settings */

	InitGraphics();

	idi->iolo = IOLO + 0;
	idi->iohi = IOLO + 20;
	idi->msglist = VideoMsgs;
}


void InitGraphics()
{
	bGrInfo.bmiHeader.bV4Size = sizeof(bGrInfo.bmiHeader);
	bGrInfo.bmiHeader.bV4Width = WIDTH;
	bGrInfo.bmiHeader.bV4Height = -HEIGHT;	//Nohama na zemi
	bGrInfo.bmiHeader.bV4Planes = 1;
	bGrInfo.bmiHeader.bV4BitCount = 1;
	bGrInfo.bmiHeader.bV4V4Compression = BI_RGB;
	bGrInfo.bmiHeader.bV4SizeImage = 0;
	bGrInfo.bmiHeader.bV4ClrUsed = 0;
	bGrInfo.bmiHeader.bV4ClrImportant = 0;
	bGrInfo.bmiHeader.bV4AlphaMask = 0;
	bGrInfo.bmiHeader.bV4CSType = LCS_WINDOWS_COLOR_SPACE;
	bGrInfo.bmiColors[0].rgbRed = 255;
	bGrInfo.bmiColors[0].rgbGreen = 255;
	bGrInfo.bmiColors[0].rgbBlue = 255;
	bGrInfo.bmiColors[1].rgbRed = 0;
	bGrInfo.bmiColors[1].rgbGreen = 0;
	bGrInfo.bmiColors[1].rgbBlue = 0;

	hOffDC = CreateCompatibleDC(GetDC(info.hWnd));
	SelectObject(hOffDC, hLogo);
}


DLLEXPORT void __cdecl OnMessage(MESSAGEDLLSTRUCT* mds)
{
	int wmId, wmEvent;
	HDC hdc;
	PAINTSTRUCT ps;
	RECT rect;

	switch (mds->message) {
	case STOPVIDEO:
		VideoOn = 0;
		Repaint();
		break;
	case CPURUNNING:
		Running = TRUE;
		Repaint();
		SetTimer(info.hWnd, MYTIMER, PERIOD, &TimerProc);
		break;
	case CPUSTOPPED:
		Running = FALSE;
		KillTimer(info.hWnd, MYTIMER);
		Repaint();
		break;
	case WM_PAINT:

		hdc = BeginPaint(info.hWnd, &ps);
		
		if (Running == TRUE) {
			if (VideoOn != 0) {
				PaintScreen(hdc);
			}
			else {
				GetClientRect(info.hWnd, &rect);
				FillRect(hdc, &rect, (HBRUSH) (COLOR_APPWORKSPACE+1));
			}
		}
		else {
			BitBlt(hdc, info.rcPaint.left, info.rcPaint.top, 256, 256, hOffDC, 0, 0, SRCCOPY);
		}
		EndPaint(info.hWnd, &ps);
		break;
	}
}

VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, 
						DWORD dwTime)
{
	if (VideoOn == 0) {
		return;
	}

	Repaint();
	(*info.memsetb)(ISVBI, 1);
	SendMessage(info.hWnd, NMIMSG, 0, 0);
}

void Repaint()
{
	InvalidateRect(info.hWnd, &(info.rcPaint), FALSE);
    UpdateWindow(info.hWnd);
}

#define IOSETGET(GVAR, SVAR) \
	if (mode == IOREAD) {	\
		*data = GVAR;	\
	}	\
	else {	\
		SVAR = *data;	\
	}

DLLEXPORT void __cdecl IOFunc(UINT adr, unsigned char* data, 
										  iomode mode)
{
	switch (adr) {
	case GRMODE:
		IOSETGET(GrMode, _GrMode);
		break;
	case LOADR:
		IOSETGET(LoAdr, _LoAdr);
		break;
	case HIADR:
		IOSETGET(HiAdr, _HiAdr);
		break;
	case BKCOLR:
		IOSETGET(bGrInfo.bmiColors[0].rgbRed, _BkColR);
		break;
	case BKCOLG:
		IOSETGET(bGrInfo.bmiColors[0].rgbGreen, _BkColG);
		break;
	case BKCOLB:
		IOSETGET(bGrInfo.bmiColors[0].rgbBlue, _BkColB);
		break;
	case FGCOLR:
		IOSETGET(bGrInfo.bmiColors[1].rgbRed, _FgColR);
		break;
	case FGCOLG:
		IOSETGET(bGrInfo.bmiColors[1].rgbGreen, _FgColG);
		break;
	case FGCOLB:
		IOSETGET(bGrInfo.bmiColors[1].rgbBlue, _FgColB);
		break;
	case LOCHR:
		IOSETGET(LoChr, _LoChr);
		break;
	case HICHR:
		IOSETGET(HiChr, _HiChr);
		break;
	case VIDEOON:
		IOSETGET(VideoOn, VideoOn);
		if (VideoOn == 0) {
				InvalidateRect(info.hWnd, &(info.rcPaint), FALSE);
				UpdateWindow(info.hWnd);
		}
		break;
	case COMMAND:
		if (mode == IOREAD) {
			*data = 0;
		}
		else {
			GrMode = _GrMode;
			LoAdr = _LoAdr;
			HiAdr = _HiAdr;
			LoChr = _LoChr;
			HiChr = _HiChr;
			memStart = ADR(HiAdr, LoAdr);
			memChrStart = ADR(HiChr, LoChr);

			bGrInfo.bmiColors[0].rgbRed = _BkColR;
			bGrInfo.bmiColors[0].rgbGreen  = _BkColG;
			bGrInfo.bmiColors[0].rgbBlue  = _BkColB;
			bGrInfo.bmiColors[1].rgbRed = _FgColR;
			bGrInfo.bmiColors[1].rgbGreen = _FgColG;
			bGrInfo.bmiColors[1].rgbBlue = _FgColB;
		}
		break;
	}
}

void PaintScreen(HDC hdc)
{
	UINT i, ic, j, jc;
	unsigned char data;

	switch(GrMode) {
	case 0:		/* TEXT MODE */
		for(i = 0, ic = 0; 	ic < XCOUNT; 
				i += FWIDTH, ic++) {
			for(j = 0, jc = 0; jc < YCOUNT; 
					j += FHEIGHT, jc++) {
				data = (*info.memgetb)(memStart + ic + jc * XCOUNT);
				CopyChar(OffScreen, memChrStart + data * 8, 
					i, j, (data < 128) ? FALSE : TRUE);
			}
		}
		SetDIBitsToDevice(hdc, info.rcPaint.left, info.rcPaint.top, WIDTH, HEIGHT, 0, 0, 0, HEIGHT, (void*) OffScreen, (BITMAPINFO*) &bGrInfo, DIB_RGB_COLORS);
		//SetDIBitsToDevice(hdc, 0, 0, WIDTH, HEIGHT, 0, 0, 0, HEIGHT, (void*) OffScreen, (BITMAPINFO*) &bGrInfo, DIB_RGB_COLORS);
		break;
	case 1:		/* GRAPHICS MODE */
		EnterCriticalSection(&info.CSMemory);
		SetDIBitsToDevice(hdc, info.rcPaint.left, info.rcPaint.top, WIDTH, HEIGHT, 0, 0, 0, HEIGHT, (void*) (info.memory + memStart), (BITMAPINFO*) &bGrInfo, DIB_RGB_COLORS);
		LeaveCriticalSection(&info.CSMemory);
		break;
	}
}

void CopyChar(char* dest, UINT adr, int x, int y, BOOL invert)
{
	int i;
	for (i = 0; i < FHEIGHT; i++) {
		if (invert == FALSE) {
			dest[(y + i) * (WIDTH / 8) + x / 8] = (*info.memgetb)(adr + i);
		}
		else {
			dest[(y + i) * (WIDTH / 8) + x / 8] = ~(*info.memgetb)(adr + i - 128 * 8);
		}
	}
}
